from __future__ import division

from builtins import bytes, str, zip
from itertools import tee, islice, chain


def from_base(digits, base):
    return sum(v * (base ** (len(digits) - k - 1)) for k, v in enumerate(digits))


def to_base(value, base):
    digits = []

    while value > 0:
        digits.insert(0, value % base)
        value //= base

    return digits


def switch_base(digits, source_base, target_base):
    return to_base(from_base(digits, source_base), target_base)


def chunks(iterable, size):
    """Generator which chunks data into chunks of given size."""
    it = iter(iterable)
    while True:
        chunk = tuple(islice(it, size))
        if not chunk:
            return
        yield chunk


def to_bytes(input, encoding='utf-8'):
    if isinstance(input, bytes):
        return bytes(input)

    if isinstance(input, str):
        return bytes(input, encoding)

    raise ValueError("Invalid input, expected string or bytes")


def iterate_prev_next(iterable):
    """
    Creates an iterator which provides previous, current and next item.
    """
    prevs, items, nexts = tee(iterable, 3)
    prevs = chain([None], prevs)
    nexts = chain(islice(nexts, 1, None), [None])
    return zip(prevs, items, nexts)



from builtins import range

from .data import ERROR_CORRECTION_FACTORS


def compute_error_correction_code_words(data_words, level):
    assert 0 <= level <= 8

    # Correction factors for the given level
    factors = ERROR_CORRECTION_FACTORS[level]

    # Number of EC words
    count = 2 ** (level + 1)

    # Correction code words list, prepopulated with zeros
    ec_words = [0] * count

    # Do the math
    for data_word in data_words:
        temp = (data_word + ec_words[-1]) % 929

        for x in range(count - 1, -1, -1):
            word = ec_words[x - 1] if x > 0 else 0
            ec_words[x] = (word + 929 - (temp * factors[x]) % 929) % 929

    return [929 - x if x > 0 else x for x in reversed(ec_words)]



"""
Text Compaction Mode (TC)

Can encode: ASCII 9, 10, 13 and 32-126
Rate compaction: 2 bytes per code word
"""

from itertools import chain
from pdf417gen.data import CHARACTERS_LOOKUP, SWITCH_CODES, Submode
from pdf417gen.util import chunks


def _exists_in_submode(char, submode):
    return char in CHARACTERS_LOOKUP and submode in CHARACTERS_LOOKUP[char]


def _get_submode(char):
    if char not in CHARACTERS_LOOKUP:
        raise ValueError("Cannot encode char: {}".format(char))

    submodes = CHARACTERS_LOOKUP[char].keys()

    preference = [Submode.LOWER, Submode.UPPER, Submode.MIXED, Submode.PUNCT]

    for submode in preference:
        if submode in submodes:
            return submode

    raise ValueError("Cannot encode char: {}".format(char))


def compact_text_interim(data):
    """Encodes text data to interim code words."""

    def _interim_text_generator(data):
        # By default, encoding starts in uppercase submode
        submode = Submode.UPPER

        for char in data:
            # Switch submode if needed
            if not _exists_in_submode(char, submode):
                prev_submode = submode
                submode = _get_submode(char)
                yield SWITCH_CODES[prev_submode][submode]

            yield [CHARACTERS_LOOKUP[char][submode]]

    return chain(*_interim_text_generator(data))


# Since each code word consists of 2 characters, a padding value is
# needed when encoding a single character. 29 is used as padding because
# it's a switch in all 4 submodes, and doesn't add any data.
PADDING_INTERIM_CODE = 29


def _compact_chunk(chunk):
    if len(chunk) == 1:
        chunk = (chunk[0], PADDING_INTERIM_CODE)

    return 30 * chunk[0] + chunk[1]


def compact_text(data):
    """Encodes data into code words using the Text compaction mode."""
    interim_codes = compact_text_interim(data)
    return (_compact_chunk(chunk) for chunk in chunks(interim_codes, 2))



from collections import namedtuple
from itertools import chain, groupby

from pdf417gen.compaction import optimizations
from pdf417gen.compaction.byte import compact_bytes
from pdf417gen.compaction.numeric import compact_numbers
from pdf417gen.compaction.text import compact_text
from pdf417gen.data import CHARACTERS_LOOKUP


# Codes for switching between compacting modes
TEXT_LATCH = 900
BYTE_LATCH = 901
BYTE_LATCH_ALT = 924
BYTE_SWITCH = 913
NUMERIC_LATCH = 902


# A chunk of barcode data with accompanying compaction function
Chunk = namedtuple("Chunk", ["data", "compact_fn"])


def compact(data):
    """Encodes given data into an array of PDF417 code words."""
    chunks = _split_to_chunks(data)
    chunks = optimizations.replace_short_numeric_chunks(chunks)
    chunks = optimizations.merge_chunks_with_same_compact_fn(chunks)
    return _compact_chunks(chunks)


def _compact_chunks(chunks):
    compacted_chunks = (
        _compact_chunk(ordinal, chunk) for ordinal, chunk in enumerate(chunks))

    return chain(*compacted_chunks)


def _compact_chunk(ordinal, chunk):
    code_words = []

    # Add the switch code if required
    add_switch_code = ordinal > 0 or chunk.compact_fn != compact_text
    if add_switch_code:
        code_words.append(get_switch_code(chunk))

    code_words.extend(chunk.compact_fn(chunk.data))

    return code_words


def _split_to_chunks(data):
    """
    Splits a string into chunks which can be compacted with the same compacting
    function.
    """
    for fn, chunk in groupby(data, key=get_optimal_compactor_fn):
        yield Chunk(list(chunk), fn)


def get_optimal_compactor_fn(char):
    if 48 <= char <= 57:
        return compact_numbers

    if char in CHARACTERS_LOOKUP:
        return compact_text

    return compact_bytes


def get_switch_code(chunk):
    if chunk.compact_fn == compact_text:
        return TEXT_LATCH

    if chunk.compact_fn == compact_bytes:
        return BYTE_LATCH_ALT if len(chunk.data) % 6 == 0 else BYTE_LATCH

    if chunk.compact_fn == compact_numbers:
        return NUMERIC_LATCH

    assert False, "Nonexistant compaction function"



from __future__ import division

import math

from pdf417gen.codes import map_code_word
from pdf417gen.compaction import compact
from pdf417gen.error_correction import compute_error_correction_code_words
from pdf417gen.util import chunks, to_bytes

START_CHARACTER = 0x1fea8
STOP_CHARACTER = 0x3fa29
PADDING_CODE_WORD = 900

# Maximum nubmer of code words which can be contained in a bar code, including
# the length descriptor, data, error correction and padding
MAX_CODE_WORDS = 928

# Limits on the number of rows and columns which can be contained in a bar code
MIN_ROWS = 3
MAX_ROWS = 90

# Encoding to use when given a string and encoding is not specified
DEFAULT_ENCODING = 'utf-8'


def encode(data, columns=6, security_level=2, encoding=DEFAULT_ENCODING):
    if columns < 1 or columns > 30:
        raise ValueError("'columns' must be between 1 and 30. Given: %r" % columns)

    if security_level < 0 or security_level > 8:
        raise ValueError("'security_level' must be between 1 and 8. Given: %r" % security_level)

    num_cols = columns  # Nomenclature

    # Prepare input
    data_bytes = to_bytes(data, encoding)

    # Convert data to code words and split into rows
    code_words = encode_high(data_bytes, num_cols, security_level)
    rows = list(chunks(code_words, num_cols))

    return list(encode_rows(rows, num_cols, security_level))


def encode_rows(rows, num_cols, security_level):
    num_rows = len(rows)

    for row_no, row_data in enumerate(rows):
        left = get_left_code_word(row_no, num_rows, num_cols, security_level)
        right = get_right_code_word(row_no, num_rows, num_cols, security_level)

        yield encode_row(row_no, row_data, left, right)


def encode_row(row_no, row_words, left, right):
    table_idx = row_no % 3

    # Convert high level code words to low level code words
    left_low = map_code_word(table_idx, left)
    right_low = map_code_word(table_idx, right)
    row_words_low = [map_code_word(table_idx, word) for word in row_words]

    return [START_CHARACTER, left_low] + row_words_low + [right_low, STOP_CHARACTER]


def encode_high(data, columns, security_level):
    """Converts the input string to high level code words.

    Including the length indicator and the error correction words, but without
    padding.
    """

    # Encode data to code words
    data_words = list(compact(data))
    data_count = len(data_words)

    # Get the padding to align data to column count
    ec_count = 2 ** (security_level + 1)
    padding_words = get_padding(data_count, ec_count, columns)
    padding_count = len(padding_words)

    # Length descriptor includes the data CWs, padding CWs and the descriptor
    # itself but not the error correction CWs
    length_descriptor = data_count + padding_count + 1

    # Total number of code words and number of rows
    cw_count = data_count + ec_count + padding_count + 1
    row_count = math.ceil(cw_count / columns)

    # Check the generated bar code's size is within specification parameters
    validate_barcode_size(length_descriptor, row_count)

    # Join encoded data with the length specifier and padding
    extendend_words = [length_descriptor] + data_words + padding_words

    # Calculate error correction words
    ec_words = compute_error_correction_code_words(extendend_words, security_level)

    return extendend_words + ec_words


def validate_barcode_size(length_descriptor, row_count):
    if length_descriptor > MAX_CODE_WORDS:
        raise ValueError(
            "Data too long. Generated bar code has length descriptor of %d. "
            "Maximum is %d." % (length_descriptor, MAX_CODE_WORDS))

    if row_count < MIN_ROWS:
        raise ValueError(
            "Generated bar code has %d rows. Minimum is %d rows. "
            "Try decreasing column count." % (row_count, MIN_ROWS))

    if row_count > MAX_ROWS:
        raise ValueError(
            "Generated bar code has %d rows. Maximum is %d rows. "
            "Try increasing column count." % (row_count, MAX_ROWS))


def get_left_code_word(row_no, num_rows, num_cols, security_level):
    table_id = row_no % 3

    if table_id == 0:
        x = (num_rows - 1) // 3
    elif table_id == 1:
        x = security_level * 3 + (num_rows - 1) % 3
    elif table_id == 2:
        x = num_cols - 1

    return 30 * (row_no // 3) + x


def get_right_code_word(row_no, num_rows, num_cols, security_level):
    table_id = row_no % 3

    if table_id == 0:
        x = num_cols - 1
    elif table_id == 1:
        x = (num_rows - 1) // 3
    elif table_id == 2:
        x = security_level * 3 + (num_rows - 1) % 3

    return 30 * (row_no // 3) + x


def get_padding(data_count, ec_count, num_cols):
    # Total number of data words and error correction words, additionally
    # reserve 1 code word for the length descriptor
    total_count = data_count + ec_count + 1
    mod = total_count % num_cols

    return [PADDING_CODE_WORD] * (num_cols - mod) if mod > 0 else []



from itertools import chain, groupby
from pdf417gen.compaction.numeric import compact_numbers
from pdf417gen.compaction.text import compact_text
from pdf417gen.util import iterate_prev_next


def replace_short_numeric_chunks(chunks):
    """
    The Numeric Compaction mode can pack almost 3 digits (2.93) into a symbol
    character. Though Numeric Compaction mode can be invoked at any digit
    length, it is recommended to use Numeric Compaction mode when there are
    more than 13 consecutive digits. Otherwise, use Text Compaction mode.
    """
    from pdf417gen.compaction import Chunk

    for prev, chunk, next in iterate_prev_next(chunks):
        is_short_numeric_chunk = (
            chunk.compact_fn == compact_numbers
            and len(chunk.data) < 13
        )

        borders_text_chunk = (
            (prev and prev.compact_fn == compact_text) or
            (next and next.compact_fn == compact_text)
        )

        if is_short_numeric_chunk and borders_text_chunk:
            yield Chunk(chunk.data, compact_text)
        else:
            yield chunk


def merge_chunks_with_same_compact_fn(chunks):
    from pdf417gen.compaction import Chunk

    for compact_fn, group in groupby(chunks, key=lambda x: x[1]):
        data = chain.from_iterable(chunk.data for chunk in group)
        yield Chunk(list(data), compact_fn)



"""
Numeric Compaction Mode (NC)

Can encode: Digits 0-9, ASCII
Rate compaction: 2.9 bytes per code word
"""

from itertools import chain
from pdf417gen.util import to_base, chunks


def _compact_chunk(chunk):
    number = "".join(chr(x) for x in chunk)
    value = int("1" + number)
    return to_base(value, 900)


def compact_numbers(data):
    """Encodes data into code words using the Numeric compaction mode."""
    compacted_chunks = (_compact_chunk(chunk) for chunk in chunks(data, size=44))
    return chain(*compacted_chunks)



"""
Byte Compaction Mode (BC)

Can encode: ASCII 0 to 255
Rate compaction: 1.2 byte per code word
"""

from itertools import chain
from pdf417gen.util import switch_base, chunks


def compact_bytes(data):
    """Encodes data into code words using the Byte compaction mode."""
    compacted_chunks = (_compact_chunk(chunk) for chunk in chunks(data, size=6))
    return chain(*compacted_chunks)


def _compact_chunk(chunk):
    """
    Chunks of exactly 6 bytes are encoded into 5 codewords by using a base 256
    to base 900 transformation. Smaller chunks are left unchanged.
    """
    digits = [i for i in chunk]

    if len(chunk) == 6:
        return switch_base(digits, 256, 900)

    return digits



import sys

from argparse import ArgumentParser

from pdf417gen import encode, render_image


def print_usage():
    print("Usage: pdf417gen [command]")
    print("")
    print("Commands:")
    print("  help    show this help message and exit")
    print("  encode  generate a bar code from given input")
    print("")
    print("https://github.com/ihabunek/pdf417gen")


def print_err(msg):
    sys.stderr.write('\033[91m' + msg + '\033[0m' + "\n")


def get_parser():
    parser = ArgumentParser(epilog="https://github.com/ihabunek/pdf417gen",
                            description="Generate a bar code from given input")

    parser.add_argument("text", type=str, nargs="?",
                        help="Text or data to encode. Alternatively data can be piped in.")

    parser.add_argument("-c", "--columns", dest="columns", type=int,
                        help="The number of columns (default is 6).",
                        default=6)

    parser.add_argument("-l", "--security-level", dest="security_level", type=int,
                        help="Security level (default is 2).",
                        default=2)

    parser.add_argument("-e", "--encoding", dest="encoding", type=str,
                        help="Character encoding used to decode input (default is utf-8).",
                        default='utf-8')

    parser.add_argument("-s", "--scale", dest="scale", type=int,
                        help="Module width in pixels (default is 3).",
                        default=3)

    parser.add_argument("-r", "--ratio", dest="ratio", type=int,
                        help="Module height to width ratio (default is 3).",
                        default=3)

    parser.add_argument("-p", "--padding", dest="padding", type=int,
                        help="Image padding in pixels (default is 20).",
                        default=20)

    parser.add_argument("-f", "--foreground-color", dest="fg_color", type=str,
                        help="Foreground color in hex (default is '#000000').",
                        default="#000000")

    parser.add_argument("-b", "--background-color", dest="bg_color", type=str,
                        help="Foreground color in hex (default is '#FFFFFF').",
                        default="#FFFFFF")

    parser.add_argument("-o", "--output", dest="output", type=str,
                        help="Target file (if not given, will just show the barcode).")

    return parser


def do_encode(args):
    args = get_parser().parse_args(args)
    text = args.text

    # If no text is given, check stdin
    if not text:
        text = sys.stdin.read()

    if not text:
        print_err("No input given")
        return

    try:
        codes = encode(
            text,
            columns=args.columns,
            security_level=args.security_level,
            encoding=args.encoding,
        )

        image = render_image(
            codes,
            scale=args.scale,
            ratio=args.ratio,
            padding=args.padding,
            fg_color=args.fg_color,
            bg_color=args.bg_color,
        )
    except Exception as e:
        print_err(str(e))
        return

    if args.output:
        image.save(args.output)
    else:
        image.show()


def main():
    command = sys.argv[1] if len(sys.argv) > 1 else None
    args = sys.argv[2:]

    if command == 'encode':
        do_encode(args)
    else:
        print_usage()



from PIL import Image, ImageColor, ImageOps
from xml.etree.ElementTree import ElementTree, Element, SubElement


def barcode_size(codes):
    """Returns the barcode size in modules."""
    num_rows = len(codes)
    num_cols = len(codes[0])

    # 17 bodules per column, last column has an additional module
    width = num_cols * 17 + 1
    height = num_rows

    return width, height


def modules(codes):
    """Iterates over codes and yields barcode moudles as (y, x) tuples."""

    for row_id, row in enumerate(codes):
        col_id = 0
        for value in row:
            for digit in format(value, 'b'):
                if digit == "1":
                    yield col_id, row_id
                col_id += 1


def parse_color(color):
    return ImageColor.getrgb(color)


def rgb_to_hex(color):
    return '#{0:02x}{1:02x}{2:02x}'.format(*color)


def render_image(codes, scale=3, ratio=3, padding=20, fg_color="#000", bg_color="#FFF"):
    width, height = barcode_size(codes)

    # Translate hex code colors to RGB tuples
    bg_color = parse_color(bg_color)
    fg_color = parse_color(fg_color)

    # Construct the image
    image = Image.new("RGB", (width, height), bg_color)

    # Draw the pixle grid
    px = image.load()
    for x, y in modules(codes):
        px[x, y] = fg_color

    # Scale and add padding
    image = image.resize((scale * width, scale * height * ratio), resample=Image.NEAREST)
    image = ImageOps.expand(image, padding, bg_color)

    return image


def render_svg(codes, scale=3, ratio=3, color="#000", description=None):
    # Barcode size in modules
    width, height = barcode_size(codes)

    # Size of each module
    scale_x = scale
    scale_y = scale * ratio

    color = rgb_to_hex(parse_color(color))

    root = Element('svg', {
        "version": "1.1",
        "xmlns": "http://www.w3.org/2000/svg",
        "width": str(width * scale_x),
        "height": str(height * scale_y),
    })

    if description:
        description_element = SubElement(root, 'description')
        description_element.text = description

    group = SubElement(root, 'g', {
        "id": "barcode",
        "fill": color,
        "stroke": "none"
    })

    # Generate the barcode modules
    for col_id, row_id in modules(codes):
        SubElement(group, 'rect', {
            "x": str(col_id * scale_x),
            "y": str(row_id * scale_y),
            "width": str(scale_x),
            "height": str(scale_y),
        })

    return ElementTree(element=root)



#!/usr/bin/env python

from setuptools import setup

with open("README.rst") as readme:
    long_description = readme.read()

setup(
    name='pdf417gen',
    version='0.7.1',
    description='PDF417 2D barcode generator for Python',
    long_description=long_description,
    author='Ivan Habunek',
    author_email='ivan.habunek@gmail.com',
    url='https://github.com/ihabunek/pdf417-py/',
    keywords='pdf417 2d barcode generator',
    license='MIT',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
    packages=[
        'pdf417gen',
        'pdf417gen.compaction'
    ],
    install_requires=[
        'Pillow>=2.0.0',
        'future'
    ],
    entry_points={
        'console_scripts': [
            'pdf417gen=pdf417gen.console:main',
        ],
    }
)
